## ----pckg_load, message=FALSE-------------------------------------------------
## Load the SVbyEye package
library(SVbyEye)

## ----read_paf, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Get PAF file to read
paf.file <- system.file("extdata", "test1.paf",
    package = "SVbyEye"
)
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file,
    include.paf.tags = TRUE, restrict.paf.tags = "cg"
)
## Example PAF table, SVbyEye expects given column names
paf.table

## Filter alignment based on size
filterPaf(paf.table = paf.table, min.align.len = 100000)

## Force to flip orientation of PAF alignments
flipPaf(paf.table = paf.table, force = TRUE)

## ----simple_miro, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Get PAF to plot
paf.file <- system.file("extdata", "test1.paf",
    package = "SVbyEye"
)
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file,
    include.paf.tags = TRUE, restrict.paf.tags = "cg"
)
## Make a plot colored by alignment directionality
plotMiro(paf.table = paf.table, color.by = "direction")

## ----miro_color, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Use custom color palette to color alignment directionality
plotMiro(
    paf.table = paf.table,
    color.palette = c("+" = "azure3", "-" = "yellow3")
)

## ----miro_idt, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Color alignments by sequence identity
plotMiro(paf.table = paf.table, color.by = "identity")

## ----miro_bin, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Plot binned alignments
plotMiro(paf.table = paf.table, binsize = 1000)

## ----add_annot, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Make a plot
plt <- plotMiro(paf.table = paf.table)
## Load target annotation file
target.annot <- system.file("extdata", "test1_target_annot.txt",
    package = "SVbyEye"
)
target.annot.df <- read.table(target.annot,
    header = TRUE, sep = "\t",
    stringsAsFactors = FALSE
)
target.annot.gr <- GenomicRanges::makeGRangesFromDataFrame(target.annot.df)
## Add target annotation as arrowhead
addAnnotation(
    ggplot.obj = plt,
    annot.gr = target.annot.gr, coordinate.space = "target"
)

## Load query annotation file
query.annot <- system.file("extdata", "test1_query_annot.txt",
    package = "SVbyEye"
)
query.annot.df <- read.table(query.annot,
    header = TRUE, sep = "\t",
    stringsAsFactors = FALSE
)
query.annot.gr <- GenomicRanges::makeGRangesFromDataFrame(query.annot.df)
## Add query annotation as rectangle
addAnnotation(
    ggplot.obj = plt,
    annot.gr = query.annot.gr, shape = "rectangle",
    coordinate.space = "query"
)

## ----lift_annot, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Lift target annotation to query and plot
target.gr <- as("target.region:19100000-19150000", "GRanges")
lifted.annot.gr <- liftRangesToAlignment(
    paf.table = paf.table,
    gr = target.gr,
    direction = "target2query"
)
plt1 <- addAnnotation(
    ggplot.obj = plt, annot.gr = target.gr, shape = "rectangle",
    coordinate.space = "target"
)
addAnnotation(
    ggplot.obj = plt1, annot.gr = lifted.annot.gr, shape = "rectangle",
    coordinate.space = "query"
)

## ----SD_annot, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Load segmental duplication annotation
sd.annot <- system.file("extdata", "test1.sd.annot.RData", package = "SVbyEye")
sd.annot.gr <- get(load(sd.annot))
## Create a custom discrete levels based on
sd.categ <- findInterval(sd.annot.gr$fracMatch, vec = c(0.95, 0.98, 0.99))
sd.categ <- dplyr::recode(sd.categ,
    "0" = "<95%", "1" = "95-98%",
    "2" = "98-99%", "3" = ">=99%"
)
sd.categ <- factor(sd.categ, levels = c("<95%", "95-98%", "98-99%", ">=99%"))
sd.annot.gr$sd.categ <- sd.categ
## Define a custom color palette
color.palette <- c(
    "<95%" = "gray72", "95-98%" = "gray47", "98-99%" = "#cccc00",
    ">=99%" = "#ff6700"
)
## Add annotation to the plot
addAnnotation(
    ggplot.obj = plt, annot.gr = sd.annot.gr, fill.by = "sd.categ",
    color.palette = color.palette, coordinate.space = "target"
)

## ----annot_label, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Give an annotation layer its own label
## Add target annotation as arrowhead
plt1 <- addAnnotation(
    ggplot.obj = plt,
    annot.gr = target.annot.gr, coordinate.space = "target",
    annotation.label = "Region"
)
## Add SD annotation
addAnnotation(
    ggplot.obj = plt1, annot.gr = sd.annot.gr, fill.by = "sd.categ",
    color.palette = color.palette, coordinate.space = "target",
    annotation.label = "SDs"
)

## ----annot_level, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Add target annotation at the increased y-axis level
addAnnotation(
    ggplot.obj = plt, annot.gr = target.annot.gr,
    coordinate.space = "target", annotation.level = 0.2
)

## Add target annotation at the zero level
addAnnotation(
    ggplot.obj = plt, annot.gr = target.annot.gr,
    coordinate.space = "target", annotation.level = 0
)

## ----annot_gene, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Create gene-like annotation
test.gr <- GenomicRanges::GRanges(
    seqnames = "target.region",
    ranges = IRanges::IRanges(
        start = c(19000000, 19030000, 19070000),
        end = c(19010000, 19050000, 19090000)
    ),
    ID = "gene1"
)
## Add single gene annotation
addAnnotation(
    ggplot.obj = plt, annot.gr = test.gr, coordinate.space = "target",
    shape = "rectangle", annotation.group = "ID", fill.by = "ID"
)

## ----add_aln, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Make a plot
plt1 <- plotMiro(paf.table = paf.table, add.alignment.arrows = FALSE)
## Highlight alignment as a filled polygon
addAlignments(
    ggplot.obj = plt1, paf.table = paf.table[3, ], fill.by = "strand",
    fill.palette = c("+" = "red")
)
## Highlight alignment as outlined polygon by dashed line
addAlignments(ggplot.obj = plt1, paf.table = paf.table[4, ], linetype = "dashed")

## ----disjoinPAF, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Disjoin PAF at target coordinates
disj.paf.table <- disjoinPafAlignments(
    paf.table = paf.table,
    coordinates = "target"
)
## Plot disjoined alignments
plotMiro(paf.table = disj.paf.table, outline.alignments = TRUE)

## ----plotDisjoinPAF, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Plot disjoined alignments
plt <- plotMiro(
    paf.table = disj.paf.table, outline.alignments = TRUE,
    add.alignment.arrows = FALSE
)
## Highlight disjoined alignments
addAlignments(
    ggplot.obj = plt, paf.table = disj.paf.table[c(4:5), ],
    linetype = "dashed"
)

## ----highlightDisjoinPAF, echo=TRUE, message=FALSE, fig.dim = c(10, 4), dpi=36, dev='png'----
## Disjoin PAF at user defined coordinates
disj.gr <- GenomicRanges::GRanges(
    seqnames = "query.region",
    ranges = IRanges::IRanges(
        start = 16300000,
        end = 16350000
    )
)
disj.paf.table <- disjoinPafAlignments(
    paf.table = paf.table,
    coordinates = "query",
    disjoin.gr = disj.gr
)
## Plot disjoined alignments
plt <- plotMiro(paf.table = disj.paf.table, outline.alignments = TRUE)
## Highlight disjoined alignments
addAlignments(
    ggplot.obj = plt, paf.table = disj.paf.table[4, ],
    linetype = "dashed"
)

## ----getSVs, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Load the data to plot
paf.file <- system.file("extdata", "test3.paf", package = "SVbyEye")
paf.table <- readPaf(
    paf.file = paf.file, include.paf.tags = TRUE,
    restrict.paf.tags = "cg"
)

## Make plot and break alignment at the position where there are deletions >=50bp
plotMiro(paf.table = paf.table, min.deletion.size = 50, highlight.sv = "outline")

## Highlight detected deletion by filled polygons
plotMiro(paf.table = paf.table, min.deletion.size = 50, highlight.sv = "fill")

## ----exportSVs, echo=TRUE, message=FALSE--------------------------------------
## Break PAF alignment at deletions >=50bp
alns <- breakPaf(paf.table = paf.table, min.deletion.size = 50)
## Print out detected deletions
alns$SVs

## ----subsetPAF, echo=TRUE, message=FALSE--------------------------------------
paf.file <- system.file("extdata", "test1.paf", package = "SVbyEye")
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file, include.paf.tags = TRUE,
    restrict.paf.tags = "cg"
)
## Subset PAF
subsetPafAlignments(
    paf.table = paf.table,
    target.region = "target.region:19050000-19200000"
)

## ----seq_heat, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Get PAF to plot
paf.file <- system.file("extdata", "test_getFASTA.paf",
    package = "SVbyEye"
)
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file,
    include.paf.tags = TRUE, restrict.paf.tags = "cg"
)
## Make a plot colored by alignment directionality
plt <- plotMiro(paf.table = paf.table, color.by = "direction")
## Get query sequence composition
fa.file <- paf.file <- system.file("extdata", "test_getFASTA_query.fasta",
    package = "SVbyEye"
)
## Calculate GC content per 5kbp bin
gc.content <- fasta2nucleotideContent(
    fasta.file = fa.file,
    binsize = 5000, nucleotide.content = "GC"
)
## Add GC content as annotation heatmap
addAnnotation(
    ggplot.obj = plt, annot.gr = gc.content, shape = "rectangle",
    fill.by = "GC_nuc.freq", coordinate.space = "query",
    annotation.label = "GCcont"
)

## ----simple_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Get PAF to plot
paf.file <- system.file("extdata", "test_ava.paf", package = "SVbyEye")
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file, include.paf.tags = TRUE,
    restrict.paf.tags = "cg"
)
## Make a plot colored by alignment directionality
plotAVA(paf.table = paf.table, color.by = "direction")

## ----extra_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Color by fraction of matched bases in each alignment
plotAVA(
    paf.table = paf.table, color.by = "identity",
    perc.identity.breaks = c(85, 90, 95)
)
## Use custom color palette to color alignment directionality
plotAVA(
    paf.table = paf.table, color.by = "direction",
    color.palette = c("+" = "azure3", "-" = "yellow3")
)
## Bin PAF alignments into user-defined bin and color them by sequence identity
plotAVA(paf.table = paf.table, binsize = 20000)

## ----ord_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Define custom sample/sequence order
seqnames.order <- c("HG00438_2", "HG01358_2", "HG02630_2", "HG03453_2")
plotAVA(
    paf.table = paf.table, color.by = "direction",
    seqnames.order = seqnames.order
)
## Only samples present in custom sample order are being plotted
seqnames.order <- c("HG00438_2", "HG01358_2", "HG03453_2")
plotAVA(
    paf.table = paf.table, color.by = "direction",
    seqnames.order = seqnames.order
)

## ----annot_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Add annotation to all-versus-all alignments ##
plt <- plotAVA(paf.table = paf.table, color.by = "direction")
annot.file <- system.file("extdata", "test_annot_ava.RData", package = "SVbyEye")
annot.gr <- get(load(annot.file))
addAnnotation(
    ggplot.obj = plt, annot.gr = annot.gr, coordinate.space = "self",
    y.label.id = "ID"
)

## ----annotAVA_level, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Add annotation to the same level as the sequence/sample alignments
plt <- plotAVA(paf.table = paf.table, color.by = "direction")
annot.file <- system.file("extdata", "test_annot_ava.RData", package = "SVbyEye")
annot.gr <- get(load(annot.file))
addAnnotation(
    ggplot.obj = plt, annot.gr = annot.gr, coordinate.space = "self",
    y.label.id = "ID", annotation.level = 0
)

## ----annotDups_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Add annotation to the same level as the sequence/sample alignments
plt <- plotAVA(paf.table = paf.table, color.by = "direction")
annot.file <- system.file("extdata", "test_annot_ava.RData", package = "SVbyEye")
annot.gr <- get(load(annot.file))
## Define color palette
colors <- setNames(annot.gr$color, annot.gr$Repeat)
## Set fill.by variable to color each range using defined color palette
addAnnotation(
    ggplot.obj = plt, annot.gr = annot.gr, fill.by = "Repeat",
    color.palette = colors, coordinate.space = "self", y.label.id = "ID",
    annotation.level = 0
)

## ----SVs_AVA, echo=TRUE, message=FALSE, fig.dim = c(8, 5), dpi=36, dev='png'----
## Break PAF alignments at insertions and deletions and highlight them by an outline
plotAVA(
    paf.table = paf.table, color.by = "direction", min.deletion.size = 1000,
    min.insertion.size = 1000, color.palette = c("+" = "azure3", "-" = "yellow3"),
    highlight.sv = "outline"
)

## Break PAF alignments at insertions and deletions and highlight them by filled polygons
plotAVA(
    paf.table = paf.table, color.by = "direction", min.deletion.size = 1000,
    min.insertion.size = 1000, color.palette = c("+" = "azure3", "-" = "yellow3"),
    highlight.sv = "fill"
)

## ----self_aln, echo=TRUE, message=FALSE, fig.dim = c(10, 2), dpi=36, dev='png'----
## Get PAF to plot
paf.file <- system.file("extdata", "test2.paf", package = "SVbyEye")
## Read in PAF
paf.table <- readPaf(
    paf.file = paf.file, include.paf.tags = TRUE,
    restrict.paf.tags = "cg"
)
## Make a plot
## Plot alignment as horizontal dotplots and color by alignment directionality
plotSelf(paf.table = paf.table, color.by = "direction", shape = "segment")

## ----self_arc, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Plot alignment as arcs and color by alignment directionality
plotSelf(paf.table = paf.table, color.by = "direction", shape = "arc")

## ----self_arrow, echo=TRUE, message=FALSE, fig.dim = c(10, 2), dpi=36, dev='png'----
## Plot alignment as arrows and color by alignment directionality
plotSelf(paf.table = paf.table, color.by = "direction", shape = "arrow")

## ----self_color, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Plot alignment as arcs and color by alignment directionality
plotSelf(
    paf.table = paf.table, color.by = "direction", shape = "arc",
    color.palette = c("+" = "azure3", "-" = "yellow3")
)

## ----self_bin, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Bin PAF alignments into user-defined bin and color them by sequence identity
plotSelf(paf.table = paf.table, binsize = 1000)

## ----self_SVs, echo=TRUE, message=FALSE, fig.dim = c(10, 3), dpi=36, dev='png'----
## Highlight structural variants within self-alignments
plotSelf(
    paf.table = paf.table, min.deletion.size = 50, min.insertion.size = 50,
    highlight.sv = "outline", shape = "arc"
)

## ----genome_plot, echo=TRUE, message=FALSE, fig.dim = c(10, 5), dpi=36, dev='png'----
## Get PAF to plot ##
paf.file <- system.file("extdata", "PTR_test.paf", package = "SVbyEye")
## Read in PAF
paf.table <- readPaf(paf.file = paf.file)
## Make a plot ##
## Color by alignment directionality
plotGenome(
    paf.table = paf.table, chromosomes = paste0("chr", c(4, 5, 9, 12)),
    chromosome.bar.width = grid::unit(2, "mm"),
    min.query.aligned.bp = 5000000
)

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

